(define-module (elmord utils iterators)
  #:use-module (oop goops)
  #:use-module (elmord utils objects)
  #:use-module (srfi srfi-41) ;; streams
  #:use-module (ice-9 control) ;; for let/ec
  )

(defclass <iterator> () procedure)
(set-object-property! <iterator> 'documentation
 "- Class: <iterator>
    (make <iterator> #:procedure PROCEDURE)

  Iterators are, essentially, procedures with internal state which return
  successive elements of a (potentially infinite) sequence.

  The <iterator> class wraps such a procedure. In principle, a simple procedure
  would suffice as an iterator, but wrapping it in a class allows writing methods
  specializing on iterators, and subclassing it.

  <iterator> has a single slot, accessed via the ‘iterator-procedure’ method.
  The slot must contain a procedure which must be called with two arguments:
  the iterator itself (which may be useful if you subclass <iterator> to add
  your own slots to keep some state), and an end-thunk. The procedure must
  either return the next element of the sequence, or call end-thunk with no
  arguments and return its result to indicate the end of the sequence. A useful
  value to pass as the end-thunk is the ‘throw-stop-iteration’ procedure, which
  raises a 'stop-iteration exception.

  The procedure ‘next!’ can be used to get the next element of an iterator.
  However, if you are going to use the iterator repeatedly, it is more efficient
  to fetch the iterator's iterator-procedure once and call it directly.
")

(define (throw-stop-iteration)
  (throw 'stop-iteration "<iterator>" "End of iteration"))

(define (vector->iterator vec)
  (let ([i 0]
        [len (vector-length vec)])
    (make <iterator>
      #:procedure (lambda (iter end-thunk)
                    (if (>= i len)
                        (end-thunk)
                      (let ([value (vector-ref vec i)])
                        (set! i (+ i 1))
                        value))))))

(define (iterator-for-each fn iter)
  (let ([iterator-proc (iterator-procedure iter)])
    (catch 'stop-iteration
      (lambda ()
        (let loop ()
          (fn (iterator-proc iter throw-stop-iteration))
          (loop)))
      (lambda exc
        *unspecified*))))

(define (iterator-map fn iter)
  (let ([iter-proc (iterator-procedure iter)])
    (make <iterator>
      #:procedure (lambda (map-iter end-thunk)
                    (let/ec return
                      (fn (iter-proc iter (lambda () (return (end-thunk))))))))))

(define (iterator-filter ok? iter)
  (let ([iter-proc (iterator-procedure iter)])
    (make <iterator>
      #:procedure
      (let ([end-mark (list 'end)])
        (lambda (filter-iter end-thunk)
          (let loop ()
            (let ([next-val (iter-proc iter (lambda () end-mark))])
              (cond [(eq? next-val end-mark) (end-thunk)]
                    [(ok? next-val) next-val]
                    [else (loop)]))))))))

(define (iterator-reduce fn init-value iter)
  (let/ec return
    (let ([iter-proc (iterator-procedure iter)])
      (let loop ([result init-value])
        (let ([next-val (iter-proc iter (lambda () (return result)))])
          (loop (fn result next-val)))))))

(define-inlinable (next! iter)
  ((iterator-procedure iter) iter throw-stop-iteration))

(define (iterator->stream iter)
  (let* ([iter-proc (iterator-procedure iter)]
         [stop-mark (list 'stop)]
         [return-stop (lambda () stop-mark)])
    (stream-let loop ()
      (let ([next-val (iter-proc iter return-stop)])
        (if (eq? next-val stop-mark)
            stream-null
          (stream-cons next-val (loop)))))))

(define (stream->iterator stream)
  (make <iterator>
    #:procedure (lambda (iter end-thunk)
                  (if (stream-null? stream)
                      (end-thunk)
                    (let ([next-val (stream-car stream)])
                      (set! stream (stream-cdr stream))
                      next-val)))))

(define (list->iterator lst)
  (make <iterator>
    #:procedure (lambda (iter end-thunk)
                  (if (null? lst)
                      (end-thunk)
                    (let ([next-val (car lst)])
                      (set! lst (cdr lst))
                      next-val)))))

(define (iterator->list iter)
  (let/ec return
    (let* ([result (list #f)]
           [tail result]
           [iter-proc (iterator-procedure iter)])
      (let loop ()
        (let ([next-val (iter-proc iter (lambda () (return (cdr result))))])
          (set-cdr! tail (list next-val))
          (set! tail (cdr tail))
          (loop))))))

;; is this faster? slower? the same?
(define (iterator->list* iter)
  (let* ([iter-proc (iterator-procedure iter)]
         [stop-mark (list 'stop)]
         [return-stop (lambda () stop-mark)])
    (let loop ([result '()])
      (let ([next-val (iter-proc iter return-stop)])
        (if (eq? next-val stop-mark)
            (reverse! result)
          (loop (cons next-val result)))))))

(define* (iterator-range start past #:optional [step 1])
  (make <iterator>
    #:procedure (let ([i start])
                  (lambda (iter end-thunk)
                    (if (< i past)
                        (let ([next-value i])
                          (set! i (+ i step))
                          next-value)
                      (end-thunk))))))
                        
