(define-module (elmord utils iterators4)
  #:use-module (oop goops)
  #:use-module (elmord utils objects)
  #:use-module (srfi srfi-41) ;; streams
  #:use-module (ice-9 control) ;; for let/ec
  )

(defclass <iterator> () procedure)
(set-object-property! <iterator> 'documentation
 "- Class: <iterator>
    (make <iterator> #:procedure PROCEDURE)

  Iterators are, essentially, procedures with internal state which return
  successive elements of a (potentially infinite) sequence.

  The <iterator> class wraps such a procedure. In principle, a simple procedure
  would suffice as an iterator, but wrapping it in a class allows writing methods
  specializing on iterators, and subclassing it.

  <iterator> has a single slot, accessed via the ‘iterator-procedure’ method.
  The slot must contain a procedure which must be called with two arguments:
  the iterator itself (which may be useful if you subclass <iterator> to add
  your own slots to keep some state), and an end-thunk. The procedure must
  either return the next element of the sequence, or call end-thunk with no
  arguments and return its result to indicate the end of the sequence. A useful
  value to pass as the end-thunk is the ‘throw-stop-iteration’ procedure, which
  raises a 'stop-iteration exception.

  The procedure ‘next!’ can be used to get the next element of an iterator.
  However, if you are going to use the iterator repeatedly, it is more efficient
  to fetch the iterator's iterator-procedure once and call it directly.
")

(define (throw-stop-iteration)
  (throw 'stop-iteration "<iterator>" "End of iteration"))

(define (vector->iterator vec)
  (let ([i 0]
        [len (vector-length vec)])
    (make <iterator>
      #:procedure (lambda (end)
                    (if (>= i len)
                        end
                      (let ([value (vector-ref vec i)])
                        (set! i (+ i 1))
                        value))))))

(define (iterator-for-each fn iter)
  (let ([iter-proc (iterator-procedure iter)]
        [end-mark (list 'end)])
    (let loop ()
      (let ([val (iter-proc end-mark)])
        (when (not (eq? val end-mark))
          (fn val)
          (loop))))))

(define (iterator-map fn iter)
  (let ([iter-proc (iterator-procedure iter)]
        [end-mark (list 'end)])
    (make <iterator>
      #:procedure (lambda (end)
                    (let ([val (iter-proc end-mark)])
                      (if (eq? val end-mark)
                          end
                        (fn val)))))))

(define (iterator-filter ok? iter)
  (let ([iter-proc (iterator-procedure iter)])
    (make <iterator>
      #:procedure
      (let ([end-mark (list 'end)])
        (lambda (end)
          (let loop ()
            (let ([val (iter-proc end-mark)])
              (cond [(eq? val end-mark) end]
                    [(ok? val) val]
                    [else (loop)]))))))))

(define (iterator-reduce fn init-value iter)
  (let ([iter-proc (iterator-procedure iter)]
        [end-mark (list 'end)])
    (let loop ([result init-value])
      (let ([val (iter-proc end-mark)])
        (if (eq? val end-mark)
            result
          (loop (fn result val)))))))

(define (list->iterator lst)
  (make <iterator>
    #:procedure (lambda (end)
                  (if (null? lst)
                      end
                    (let ([next-val (car lst)])
                      (set! lst (cdr lst))
                      next-val)))))

(define* (iterator-range start past #:optional [step 1])
  (make <iterator>
    #:procedure (let ([i start])
                  (lambda (end)
                    (if (< i past)
                        (let ([next-value i])
                          (set! i (+ i step))
                          next-value)
                      end)))))
                        
