(define-module (elmord utils objects)
  #:use-module (oop goops)
  #:use-module (srfi srfi-88) ;; `foo:' style keywords
  #:export (defclass <simple-object>))

(define-class <simple-object> ())
(set-object-property! <simple-object> 'documentation
  "Superclass for classes created with `defclass', which see.")

(define class-canonical-module (make-object-property))

(define (unspecified)
  "Return the unspecified value."
  (if #f #t))

;; (define-syntax defclass
;;   (syntax-rules ()
;;     [(_ name (constructor fields ...))
;;      (begin
;;        (define-class name (<simple-object>) fields ...)
;;        (define (constructor fields ...)
;;          (let ([obj (make name)])
;;            (begin (slot-set! obj 'fields fields) ...)
;;            obj))
;;        (set! (class-canonical-module name) (current-module)))]))
  
;; (defclass <person> (make-person name age))

(define (class-bare-name name)
  "Given a symbol of the form ‘<foo>’, return ‘foo’."
  (let* ([str (symbol->string name)]
         [len (string-length str)])
    (when (not (and (string-prefix? "<" str)
                    (string-suffix? ">" str)))
      (error "defstruct class name must be surrounded in <...>, got:" str))
    (string->symbol (substring str 1 (- len 1)))))

(define-syntax defclass
  (lambda (form)
    "- Syntax: defclass <NAME> (OPTIONS ...) SLOT ...
    Define a class named <NAME>, with the specified SLOTs.
    <NAME> must be a a symbol surrounded with <...>.

    For each slot, an accessor names NAME-SLOT is created.

    The defined class inherits from <simple-object>, which defines a default
    printed representation for the class of the form

        ,[make <NAME> slot1: val1 ...]

    The idea is that you can quasiquote (`) a printed s-expression containing
    objects and evaluate it to construct the objects.

    OPTIONS is currently ignored."
    (syntax-case form ()
      [(_ name options slots ...)
       (let ([bare-name (class-bare-name (syntax->datum #'name))])
         #`(begin
             ;; define-class is unhygienic apparently, we need to give a fully
             ;; qualified binding in the superclass list.
             (define-class name ((@@ (elmord utils objects) <simple-object>))
               #,@(syntax-map
                   (lambda (slot)
                     (simple-object-expand-slot-description form bare-name slot))
                   #'(slots ...)))
             (set! (class-canonical-module name) (current-module))
             (unspecified)
             )
         )])))

(define (syntax-map fn syntax-list)
  (syntax-case syntax-list ()
    [() syntax-list]
    [(first . rest)
     #`(#,(fn #'first) . #,(syntax-map fn #'rest))]))

(define (simple-object-expand-slot-description form bare-name slot)
  (define (syn x) (datum->syntax form x))
  (let* ([slot-name (syntax->datum slot)]
         [slot-keyword (symbol->keyword slot-name)]
         [slot-accessor (symbol-append bare-name '- slot-name)]
         )
    #`(#,(syn slot-name)
       init-keyword: #,(syn slot-keyword)
       accessor: #,(syn slot-accessor))))
  

(define-method (write (obj <simple-object>) port)
  (let* ([class (class-of obj)]
         [class-module (class-canonical-module class)]
         [class-descriptor
          (if (eq? class (module-ref (current-module) (class-name class) #f))
              (class-name class)
            (list '@@ (module-name class-module) (class-name class)))])
    (format port ",[make ~a" class-descriptor)
    (for-each (lambda (slot-descriptor)
                (when (slot-bound? obj (slot-definition-name slot-descriptor))
                  (let* ([slot-name (slot-definition-name slot-descriptor)]
                         [slot-value (slot-ref obj slot-name)])
                    (format port " ~s" (symbol->keyword slot-name))
                    (if (self-evaluating? slot-value)
                        (format port " ~s" slot-value)
                      (format port " `~s" slot-value)))))
              (class-slots class))
    (format port "]")))
                

;; Preliminary attempt at a "defclass for records". The idea is to be
;; able to use the same syntax of defclass but for defining records,
;; which are less flexible but have better performance. (Although now
;; I wonder if the same thing could be achieved with a record-like
;; metaclass instead.) The code below is kinda terrible (not that the
;; code above isn't, but anyway).

(define-syntax defrecord
  (lambda (form)
    (define (syn x) (datum->syntax form x))
    (syntax-case form ()
      [(_ name options slots ...)
       (let* ([bare-name (class-bare-name (syntax->datum #'name))]
              [record-type-var (syn (symbol-append '% (syntax->datum #'name)))]
              [%make-name (symbol-append '%make- bare-name)]
              [make-name (symbol-append 'make- bare-name)]
              [name? (symbol-append bare-name '?)])
         #`(begin
             (define #,record-type-var (make-record-type '#,(syn bare-name) '(slots ...)))
             (define #,(syn %make-name) (record-constructor #,record-type-var))
             (define* (#,(syn make-name) #:key slots ...)
               (#,(syn %make-name) slots ...))
             (define #,(syn name?) (record-predicate #,record-type-var))
             #,@(syntax-map (lambda (slot) (defstruct-make-accessor form record-type-var bare-name slot))
                            #'(slots ...))
             (define name (record-type-class #,record-type-var))))])))

(define (defstruct-make-accessor form record-type-var bare-name slot)
  (define (syn x) (datum->syntax form x))
  (let ([accessor-name (symbol-append bare-name '- (syntax->datum slot))])
    #`(define #,(syn accessor-name)
        (make-procedure-with-setter (record-accessor #,record-type-var '#,slot)
                                    (record-modifier #,record-type-var '#,slot)))))
         
(define (record-type-class rtd)
  (let ([instance (apply (record-constructor rtd)
                         (record-type-fields rtd))])
    (class-of instance)))
