(define-module (elmord utils sequences)
  #:use-module (oop goops)
  #:use-module (elmord utils objects)
  #:use-module (srfi srfi-1)  ;; list library
  #:use-module (srfi srfi-41) ;; streams
  #:export (at at-getter at-setter defaulting
            first rest empty?
            vector->stream stream->vector)
  )

;;;; Generic 'at' procedure for indexing various kinds of datatypes.

(define-generic at-getter)
(define-generic at-setter)

(define at (make-procedure-with-setter at-getter at-setter))

(set-procedure-property! at 'documentation
  "- Procedure with setter: at OBJECT INDEX ...
    Generic procedure for indexing various kinds of datatypes.

    Examples:
      (define lst (list 'foo 'bar 'baz))
      (at lst 2)                 ==> baz
      (set! (at lst 2) 'quux)
      lst                        ==> (foo bar quux)

    This procedure invokes the generic procedures at-getter and at-setter to
    perform its work. You can add new methods to those procedures to extend
    the kinds of objects that 'at' can be applied to.")

(define-method (at-getter (vec <vector>) index)
  (vector-ref vec index))

(define-method (at-setter (vec <vector>) index value)
  (vector-set! vec index value))

(define-method (at-getter (str <string>) index)
  (string-ref str index))

(define-method (at-setter (str <string>) index value)
  (string-set! str index value))

(define-method (at-getter (lst <list>) index)
  (list-ref lst index))

(define-method (at-setter (lst <list>) index value)
  (list-set! lst index value))

(define-method (at-getter (tbl <hashtable>) key)
  (let ([handle (hash-get-handle tbl key)])
    (if handle
        (cdr handle)
      (throw 'out-of-range "at" "Key not present in hashtable: ~s"
             (list key) (list key)))))

(define-method (at-setter (tbl <hashtable>) key value)
  (hash-set! tbl key value))

(define-syntax defaulting
  (syntax-rules ()
    "- Syntax: defaulting EXPR DEFAULT
    Evaluate EXPR. If an 'out-of-range exception is raised, evaluate DEFAULT.

    Example: (defaulting (at some-hash-table 'foo) 42)"
    [(defaulting expr default)
     (catch 'out-of-range
       (lambda () expr)
       (lambda args default))]))


;;;; Iteration over sequences.

(define-generic first)
(define-generic rest)
(define-generic empty?)

;;; Lists.

(define-method (first (lst <list>))
  (car lst))

(define-method (rest (lst <list>))
  (cdr lst))

(define-method (empty? (lst <list>))
  (null? lst))

;;; Vectors.

(define-method (first (vec <vector>))
  (vector-ref vec 0))

(define-method (rest (vec <vector>))
  (make <sequence-bookmark>
    sequence: vec
    index: 1
    ref: vector-ref
    length: vector-length))

(define-method (empty? (vec <vector>))
  (zero? (vector-length vec)))

;;; Strings.

(define-method (first (str <string>))
  (string-ref str 0))

(define-method (rest (str <string>))
  (make <sequence-bookmark>
    sequence: str
    index: 1
    ref: string-ref
    length: string-length))

(define-method (empty? (str <string>))
  (zero? (string-length str)))

;;; Sequence bookmarks.
;; These are used to keep track of the position within a sequence.
;; For instance, (rest vector) does not return a new vector missing
;; the first element, but rather a sequence-bookmark pointing to the
;; next vector index.

(defclass <sequence-bookmark> ()
  sequence index ref length)

(define-method (first (bk <sequence-bookmark>))
  ((sequence-bookmark-ref bk) (sequence-bookmark-sequence bk)
                              (sequence-bookmark-index bk)))

(define-method (rest (bk <sequence-bookmark>))
  (let* ([seq (sequence-bookmark-sequence bk)]
         [index (sequence-bookmark-index bk)]
         [len ((sequence-bookmark-length bk) seq)])
    (if (>= index len)
        (throw 'out-of-range "rest" "Reached end of sequence: ~s"
               (list seq) (list seq))
      (make <sequence-bookmark> sequence: seq
                                index: (+ index 1)
                                ref: (sequence-bookmark-ref bk)
                                length: (sequence-bookmark-length bk)))))

(define-method (empty? (bk <sequence-bookmark>))
  (>= (sequence-bookmark-index bk)
      ((sequence-bookmark-length bk)
       (sequence-bookmark-sequence bk))))
        
;;; Streams.

(define-method (first (stream <stream-promise>))
  (stream-car stream))

(define-method (rest (stream <stream-promise>))
  (stream-cdr stream))

(define-method (empty? (stream <stream-promise>))
  (stream-null? stream))


(define* (vector->stream vec #:optional [start 0] [end (vector-length vec)])
  "- Procedure: vector->stream VECTOR #:optional START END
  Return a stream of the successive elements of VECTOR, from position START
  (inclusive) to END (exclusive).

  If START is not provided, it defaults to 0.

  If END is not provided, it defaults to the length of VECTOR."
  (stream-let loop ([i start])
    (if (>= i end)
        stream-null
      (stream-cons (vector-ref vec i)
                   (loop (+ i 1))))))

(define (stream->vector stream)
  "- Procedure: stream->vector STREAM
  Return a vector of the elements of STREAM. STREAM should be a finite stream;
  otherwise, this function will enter an infinite loop."
  (list->vector (stream->list stream))) ;; got lazy here.
